/**************************************************************************
 *
 * backend.c -- Underlying routines for mgquery
 * Copyright (C) 1994  Neil Sharman
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: backend.cpp 711 1999-10-17 23:43:31Z cs025 $
 *
 **************************************************************************/

#include "sysfuncs.h"

#include "memlib.h"
#include "messages.h"
#include "timing.h"
#include "filestats.h"
#include "sptree.h"
#include "netorder.h"  /* [RPAP - Jan 97: Endian Ordering] */


#include "mg_files.h"
#include "mg.h"
#include "invf.h"
#include "text.h"
#include "lists.h"
#include "backend.h"
#include "stem_search.h"
#include "StemBlock.h"
#include "invf_get.h"
#include "text_get.h"
#include "weights.h"
#include "locallib.h"
#include "mg_errors.h"
#include "DocEntry.h"

static File *
OpenFile (char *base, char *suffix, unsigned long magic, int *ok)
{
  char FileName[512];
  File *F;
  sprintf (FileName, "%s%s", base, suffix);
  if (!(F = Fopen (FileName, "rb", 0)))  /* [RPAP - Feb 97: WIN32 Port] */
    {
      mg_errno = MG_NOFILE;
      MgErrorData (FileName);
      if (ok)
	*ok = 0;
      return (NULL);
    }
  if (magic)
    {
      unsigned long m;
      if (fread ((char *) &m, sizeof (m), 1, F->f) == 0)
	{
	  mg_errno = MG_READERR;
	  MgErrorData (FileName);
	  if (ok)
	    *ok = 0;
	  Fclose (F);
	  return (NULL);
	}
      NTOHUL(m);  /* [RPAP - Jan 97: Endian Ordering] */
      if (m != magic)
	{
	  mg_errno = MG_BADMAGIC;
	  MgErrorData (FileName);
	  if (ok)
	    *ok = 0;
	  Fclose (F);
	  return (NULL);
	}
    }
  return (F);
}

/* --
   -- GRB: Open text files only; based on open_all_files in MG1.2.1; 
   --      should only be called after open_index_files
   --
*/
static int 
open_text_files (query_data * qd)
{
  int ok = 1;

  qd->File_text = OpenFile (qd->textpathname, TEXT_SUFFIX, /* [RJM 06/97: text filename] */
			    MAGIC_TEXT, &ok);
  qd->File_fast_comp_dict = OpenFile (qd->textpathname, /* [RJM 06/97: text filename] */
				      TEXT_DICT_FAST_SUFFIX, MAGIC_FAST_DICT, NULL);
  if (!qd->File_fast_comp_dict)
    {
      qd->File_comp_dict = OpenFile (qd->textpathname, /* [RJM 06/97: text filename] */
				     TEXT_DICT_SUFFIX, MAGIC_DICT, &ok);
      qd->File_aux_dict = OpenFile (qd->textpathname, /* [RJM 06/97: text filename] */
				    TEXT_DICT_AUX_SUFFIX, MAGIC_AUX_DICT, NULL);
    }
  else
    qd->File_comp_dict = qd->File_aux_dict = NULL;

  /* This will fail if a level 1 inverted file was created because there 
     will be no document weights */
  qd->File_text_idx_wgt = OpenFile (qd->pathname, TEXT_IDX_WGT_SUFFIX,
				    MAGIC_TEXI_WGT, NULL);
  if (qd->File_text_idx_wgt == NULL && qd->File_weight_approx == NULL)
    qd->File_text_idx = OpenFile (qd->textpathname, /* [RJM 06/97: text filename] */
				  TEXT_IDX_SUFFIX, MAGIC_TEXI, NULL);
  else
    qd->File_text_idx = NULL;

  /**
   * GRB: Additional check; if we didn't get idx_wgt and we are on
   *        a level 2 or later inversion fileset, then error.
   *      As the id structure may not have been initialised, only
   *        error if it has been; we'll assume that if it isn't
   *        initialised yet then something bespoke is happening and
   *        the programmer should deal with it; 
   *      open_all_files() calls this fn after open_index_files,
   *        in which case id is NULL if when open_all_files is called
   *        inside __InitQuerySystem (ie. when MG is just run as normal).
   *      Programmers doing their own thing should prepare their
   *        code to cope with the 2-stage initialisation of level 2
   *        or higher indexes.  The idx_wgt index is only used in
   *        retrieving text; it has no effect on indexed searches.  
   */
  if (qd->id != NULL &&
      ((qd->allfiles == 1 && qd->File_text_idx_wgt == NULL) || 
       qd->File_weight_approx == NULL) &&
      qd->id->ifh.InvfLevel >= 2)
    { ok = 0;
    }


  if (!ok)
    { if (qd->File_text)
        Fclose (qd->File_text);
      if (qd->File_fast_comp_dict)
	Fclose (qd->File_fast_comp_dict);
      if (qd->File_aux_dict)
	Fclose (qd->File_aux_dict);
      if (qd->File_comp_dict)
	Fclose (qd->File_comp_dict);

      if (qd->File_text_idx_wgt)
	Fclose (qd->File_text_idx_wgt);
      if (qd->File_text_idx)
	Fclose (qd->File_text_idx);
      return (-1);
    }
  return (0);

}

/* --
   -- GRB: Open index files only; based on open_all_files in MG1.2.1
   --
*/
static int open_index_files(query_data *qd)
{
  int ok = 1;

  qd->File_text           = NULL;
  qd->File_fast_comp_dict = NULL;
  qd->File_comp_dict      = NULL;
  qd->File_aux_dict       = NULL;

  qd->File_stem = OpenFile (qd->pathname, INVF_DICT_BLOCKED_SUFFIX,
			    MAGIC_STEM, &ok);

  /* [RPAP - Jan 97: Stem Index Change]
     These will fail if collection not built with stem indexes */
  qd->File_stem1 = OpenFile (qd->pathname, INVF_DICT_BLOCKED_1_SUFFIX,
			     MAGIC_STEM_1, NULL);
  qd->File_stem2 = OpenFile (qd->pathname, INVF_DICT_BLOCKED_2_SUFFIX,
			     MAGIC_STEM_2, NULL);
  qd->File_stem3 = OpenFile (qd->pathname, INVF_DICT_BLOCKED_3_SUFFIX,
			     MAGIC_STEM_3, NULL);

  qd->File_invf = OpenFile (qd->pathname, INVF_SUFFIX,
			    MAGIC_INVF, &ok);

  /* Ths will fail if a level 1 inverted file was created because there 
     will be no document weights */
  qd->File_weight_approx = OpenFile (qd->pathname, APPROX_WEIGHTS_SUFFIX,
				     MAGIC_WGHT_APPROX, NULL);

  qd->File_text_idx_wgt = NULL;
  qd->File_text_idx = NULL;

  if (!ok)
    { Fclose (qd->File_stem);

      /* [RPAP - Jan 97: Stem Index Change] */
      if (qd->File_stem1)
	Fclose (qd->File_stem1);
      if (qd->File_stem2)
	Fclose (qd->File_stem2);
      if (qd->File_stem3)
	Fclose (qd->File_stem3);

      Fclose (qd->File_invf);
      if (qd->File_weight_approx)
	Fclose (qd->File_weight_approx);
      return (-1);
    }
  return (0);

}

static int 
open_all_files (query_data * qd)
{
  int ok = 1;

  if (open_index_files(qd) != 0)
    {
      return -1;
    }
  if (open_text_files(qd) != 0)
    {
      return -1;
    }
  return 0;
}

static void 
close_all_files (query_data * qd)
{ if (qd->File_text)
    Fclose (qd->File_text);
  if (qd->File_fast_comp_dict)
    Fclose (qd->File_fast_comp_dict);
  if (qd->File_aux_dict)
    Fclose (qd->File_aux_dict);
  if (qd->File_comp_dict)
    Fclose (qd->File_comp_dict);
  Fclose (qd->File_stem);

  /* [RPAP - Jan 97: Stem Index Change] */
  if (qd->File_stem1)
    Fclose (qd->File_stem1);
  if (qd->File_stem2)
    Fclose (qd->File_stem2);
  if (qd->File_stem3)
    Fclose (qd->File_stem3);

  Fclose (qd->File_invf);
  if (qd->File_text_idx_wgt)
    Fclose (qd->File_text_idx_wgt);
  if (qd->File_weight_approx)
    Fclose (qd->File_weight_approx);
  if (qd->File_text_idx)
    Fclose (qd->File_text_idx);
}

/* If textname is equal to null then name will be used instead */
/* [RJM 06/97: text filename] */

/**
 * The following define and the variable abortvar in the following function are designed
 * to provide an exception-like recovery system for when an error occurs.
 */
#define abort(x) abortvar=x;goto Recovery;

query_data *
__InitQuerySystem (char *dir, char *name, char *textname, InitQueryTimes * iqt, int allfiles)
{
  query_data *qd;
  char *s;
  int abortvar;

  if (textname == NULL) textname = name; /* [RJM 06/97: text filename] */

  if (!(qd = new query_data))
    {
      abort(0);
    }

  bzero ((char *) qd, sizeof (*qd));

  /* -- GRB: 13/09/99: note if all files were requested */
  qd->allfiles = allfiles;

  qd->mem_in_use = qd->max_mem_in_use = 0;

  qd->doc_pos = qd->buf_in_use = 0;
  qd->TextBufferLen = 0;
  qd->DL = NULL;

  /* [RPAP - Feb 97: Term Frequency] */
  qd->TL = NULL;
  qd->QTL = NULL;

  qd->TextBuffer = NULL;

  qd->tot_hops_taken = 0;
  qd->tot_num_of_ptrs = 0;
  qd->tot_num_of_accum = 0;
  qd->tot_num_of_terms = 0;
  qd->tot_num_of_ans = 0;
  qd->tot_text_idx_lookups = 0;

  qd->hops_taken = 0;
  qd->num_of_ptrs = 0;
  qd->num_of_accum = 0;
  qd->num_of_terms = 0;
  qd->num_of_ans = 0;
  qd->text_idx_lookups = 0;

  qd->pathname = NULL; /* RJM 06/97: text filename] */
  qd->textpathname = NULL; /* RJM 06/97: text filename] */

  s = strrchr (dir, '/');
  if (s && *(s + 1) == '\0')
    {
      /* [RJM 06/97: text filename] */
      if (!(qd->pathname = new char[strlen (dir) + strlen (name) + 1]) ||
	  !(qd->textpathname = new char[strlen (dir) + strlen (textname) + 1]))
	{
	  abort(1);
	}
      sprintf (qd->pathname, "%s%s", dir, name);
      sprintf (qd->textpathname, "%s%s", dir, textname); /* [RJM 06/97: text filename] */
    }

  else
    {
      /* [RJM 06/97: text filename] */
      if (!(qd->pathname = new char[strlen (dir) + strlen (name) + 2]) ||
	  !(qd->textpathname = new char[strlen (dir) + strlen (textname) + 2]))
	{
	  abort(1);
	}
/* [RPAP - Feb 97: WIN32 Port] */
#ifdef __WIN32__
      if (dir == NULL || dir[0] == '\0') {
	sprintf (qd->pathname, "%s", name);
	sprintf (qd->textpathname, "%s", textname); /* [RJM 06/97: text filename] */
      } else {
	sprintf (qd->pathname, "%s%s", dir, name);
	sprintf (qd->textpathname, "%s%s", dir, textname); /* [RJM 06/97: text filename] */
      }
#else
      sprintf (qd->pathname, "%s/%s", dir, name);
      sprintf (qd->textpathname, "%s/%s", dir, textname); /* [RJM 06/97: text filename] */
#endif
    }

  if (((allfiles == 1) && (open_all_files (qd) == -1)) ||
      ((allfiles == 0) && (open_index_files (qd) == -1)))
    {
      abort(2);
      return (NULL);
    }

  if (iqt)
    GetTime (&iqt->Start);

  /* Initialise the stemmed dictionary system */
  if (!(qd->sd = ReadStemDictBlk (qd->File_stem, _MGErr)))
    {
      abort(3);
    }

  /* [RPAP - Jan 97: Stem Index Change] */
  if ((qd->sd->sdh.indexed & 7) && qd->File_stem1 && qd->File_stem2 && qd->File_stem3)
    {
      if (!(qd->sd->stem1 = ReadStemIdxBlk (qd->File_stem1)))
	{
	  abort(4);
	}
      if (!(qd->sd->stem2 = ReadStemIdxBlk (qd->File_stem2)))
	{
	  abort(4);
	}
      if (!(qd->sd->stem3 = ReadStemIdxBlk (qd->File_stem3)))
	{
	  abort(4);
	}
      }
  else if (qd->sd->sdh.indexed != 0)
    {
      abort(4);
    }
  else
    {
      if (qd->File_stem1)
	Fclose (qd->File_stem1);
      if (qd->File_stem2)
	Fclose (qd->File_stem2);
      if (qd->File_stem3)
	Fclose (qd->File_stem3);
      qd->File_stem1 = NULL;
      qd->File_stem2 = NULL;
      qd->File_stem3 = NULL;
      qd->sd->stem1 = NULL;
      qd->sd->stem2 = NULL;
      qd->sd->stem3 = NULL;
    }

  if (iqt)
    GetTime (&iqt->StemDict);
  if (qd->File_weight_approx)
    {
      if (!(qd->awd = LoadDocWeights (qd->File_weight_approx,
				      qd->sd->sdh.num_of_docs)))
	{
	  abort(4);
	}
    }
  else
    qd->awd = NULL;


  if (iqt)
    GetTime (&iqt->ApproxWeights);

  if ((allfiles == 1) &&
      !(qd->cd = LoadCompDict (qd->File_comp_dict, qd->File_aux_dict,
			       qd->File_fast_comp_dict)))
    {
      abort(5);
    }
  else if (allfiles == 0)
    {
      qd->cd = NULL;
    }

  if (iqt)
    GetTime (&iqt->CompDict);

  if (!(qd->id = InitInvfFile (qd->File_invf, qd->sd)))
    {
      abort(6);
    }
  if (((allfiles == 1 && qd->File_text_idx_wgt == NULL) || 
       qd->File_weight_approx == NULL) &&
      qd->id->ifh.InvfLevel >= 2)
    {
      abort(7);
    }
  if (iqt)
    GetTime (&iqt->Invf);

  if ((allfiles == 1) &&
      !(qd->td = LoadTextData (qd->File_text, qd->File_text_idx_wgt,
			       qd->File_text_idx)))
    {
      abort(8);
    }
  else if (allfiles == 0)
    {
      qd->td = NULL;
    }

/* [RPAP - Feb 97: NZDL Additions] */
#if defined(PARADOCNUM) || defined(NZDL)

/*

This code is based on the TREC_MODE code below to read the .paragraph
file to determine what document numbers correspond to what paragraphs.
This code is more space efficient, reading in the .paragraph file
into memory as an accumulate docnum array.  Eg.  the .paragraph may contain

	[5 3 6 4 7 9 4]

indicating the first document has 5 paragraphs, the next 3, etc.
This will be stored in memory as

	[0 5 8 14 18 25 34 38]

so a binary search can be performed.  The first 0 is for convenience;
it prevents testing boundary conditions.


The TREC_MODE code does this differently; it stores the array

	[1 1 1 1 1 2 2 2 3 3 3 3 3 3 ....]

allowing directy paragraph to docnum conversion, at the expense
of memory.

*/ 
  qd->paragraph = NULL;

  if ((allfiles == 1) && (qd->id->ifh.InvfLevel == 3))
    {
      unsigned long magic;
      FILE *paragraph;
      int i;
      char paraFile[512];

      sprintf(paraFile, "%s%s", qd->pathname, INVF_PARAGRAPH_SUFFIX);
      paragraph = fopen(paraFile, "rb");
      if (!paragraph)
	FatalError(1, "Unable to open 'paraFile'.", paraFile);

      fread((void *)&magic, sizeof(magic), 1, paragraph);
      qd->paragraph = new int[qd->td->cth.num_of_docs+1];
      qd->paragraph[0] = 0;
      for (i = 1; i <= qd->td->cth.num_of_docs; i++)
	{
	  int count;

	  if (fread((void *)&count, sizeof(count), 1, paragraph) != 1)
	    FatalError(1, "Unexpected EOF while reading '%s'.", paraFile);
	  NTOHSI(count);  /* [RPAP - Jan 97: Endian Ordering] */
	  qd->paragraph[i] = qd->paragraph[i-1]+count;
	}

      fclose (paragraph); /* [RJM 07/98: Memory Leak] */
    }
  
  
#endif

#ifdef TREC_MODE
  {
    extern char *trec_ids;
    extern long *trec_paras;
    int size;
    char FileName[512];
    FILE *f;
    if (!strstr (qd->pathname, "trec"))
      goto error;
    sprintf (FileName, "%s%s", qd->pathname, ".DOCIDS");
    if (!(f = fopen (FileName, "rb")))  /* [RPAP - Feb 97: WIN32 Port] */
      {
	Message ("Unable to open \"%s\"", FileName);
	goto error;
      }
    fseek (f, 0, 2);
    size = ftell (f);
    fseek (f, 0, 0);
    trec_ids = new char[size];
    if (!trec_ids)
      {
	fclose (f);
	goto error;
      }
    fread (trec_ids, 1, size, f);
    fclose (f);
    if ((allfiles == 1) && (qd->id->ifh.InvfLevel == 3))
      {
	int i, d;
	unsigned long magic;
	trec_paras = new long[qd->sd->sdh.num_of_docs];
	if (!trec_paras)
	  {
	    delete trec_ids;
	    trec_ids = NULL;
	    goto error;
	  }
	sprintf (FileName, "%s%s", qd->pathname, INVF_PARAGRAPH_SUFFIX);
	if (!(f = fopen (FileName, "rb")))  /* [RPAP - Feb 97: WIN32 Port] */
	  {
	    Message ("Unable to open \"%s\"", FileName);
	    goto error;
	  }
	if (fread ((char *) &magic, sizeof (magic), 1, f) != 1 ||
	    NTOHUL(magic) != MAGIC_PARAGRAPH)  /* [RPAP - Jan 97: Endian Ordering] */
	  {
	    fclose (f);
	    Message ("Bad magic number in \"%s\"", FileName);
	    goto error;
	  }

	for (d = i = 0; i < qd->td->cth.num_of_docs; i++)
	  {
	    int count;
	    if (fread ((char *) &count, sizeof (count), 1, f) != 1)
	      {
		fclose (f);
		goto error;
	      }
	    NTOHSI(count);  /* [RPAP - Jan 97: Endian Ordering] */
	    while (count--)
	      trec_paras[d++] = i;
	  }
	fclose (f);
      }
    goto ok;
  error:
    if (trec_ids)
      delete trec_ids;
    if (trec_paras)
      delete (trec_paras);
    trec_ids = NULL;
    trec_paras = NULL;
  ok:
    ;
  }
#endif

  if (iqt)
    GetTime (&iqt->Text);

  return (qd);

 Recovery:
  switch (abortvar)
    {
    case 8: // NB: 8 indicates a failure after the loading of the inverted file
    case 7: // NB: 7 indicates a failure during the loading of the inverted file
      FreeInvfData (qd->id);
    case 6:
      if (qd->cd)
	FreeCompDict (qd->cd);
    case 5:
      if (qd->awd)
	{
	  FreeWeights (qd->awd);
	}
    case 4:
      FreeStemDict (qd->sd);
    case 3:
      close_all_files (qd);
    case 2:
      delete (qd->textpathname); /* [RJM 06/97: text filename] */
    case 1:
      if (qd->pathname) 
	delete (qd->pathname); /* [RJM 06/97: text filename] */
      delete qd;
      break;
    }

  /* Set mg error status for particular failures; NB not every failure results in an
   * mg error value being set.
   */
  if (abortvar < 2)
    mg_errno = MG_NOMEM;
  else if (abortvar == 7)
    mg_errno = MG_INVERSION;
  return (NULL);
}



query_data *
InitQuerySystem (char *dir, char *name, char *textname, InitQueryTimes * iqt)
{ return __InitQuerySystem(dir, name, textname, iqt, 1);
}

// This function initialises the query system without loading the text dictionary etc.
query_data *
InitQuerySystemNT (char *dir, char *name, char *textname, InitQueryTimes * iqt)
{ return __InitQuerySystem(dir, name, textname, iqt, 0);
}

/*
 * Change the amount of memory currently in use
 *
 */
void 
ChangeMemInUse (query_data * qd, long delta)
{
  qd->mem_in_use += delta;
  if (qd->mem_in_use > qd->max_mem_in_use)
    qd->max_mem_in_use = qd->mem_in_use;
}


void 
FinishQuerySystem (query_data * qd)
{
/* [RJM 07/98: Memory Leak] */
#if defined(PARADOCNUM) || defined(NZDL)
  if (qd->paragraph != NULL) {
    delete (qd->paragraph);
    qd->paragraph = NULL;
  }
#endif

  if (qd->td != NULL)
    {
      FreeTextData (qd->td);
    }
  FreeInvfData (qd->id);
  if (qd->cd)
    FreeCompDict (qd->cd);

  if (qd->awd)
    FreeWeights (qd->awd);
  FreeStemDict (qd->sd);
  close_all_files (qd);
  delete (qd->textpathname); /* [RJM 06/97: text filename] */
  delete (qd->pathname);
  QueryData_FreeQueryDocs (qd);
  if (qd->TL != NULL) TermList_destroy(&qd->TL);         /* [RJM 07/98: Memory Leak] */
  if (qd->QTL != NULL) QueryTermList_free(&qd->QTL);  /* [RJM 07/98: Memory Leak] */
  delete qd;

  /* other global stuff hanging around */
  MgErrorDeinit ();
}


void 
ResetFileStats (query_data * qd)
{
  if (qd->File_text)
    ZeroFileStats (qd->File_text);
  if (qd->File_comp_dict)
    ZeroFileStats (qd->File_comp_dict);
  if (qd->File_fast_comp_dict)
    ZeroFileStats (qd->File_fast_comp_dict);
  ZeroFileStats (qd->File_stem);

  /* [RPAP - Jan 97: Stem Index Change] */
  if (qd->File_stem1)
    ZeroFileStats (qd->File_stem1);
  if (qd->File_stem2)
    ZeroFileStats (qd->File_stem2);
  if (qd->File_stem3)
    ZeroFileStats (qd->File_stem3);

  ZeroFileStats (qd->File_invf);
  if (qd->File_text_idx_wgt)
    ZeroFileStats (qd->File_text_idx_wgt);
  if (qd->File_weight_approx)
    ZeroFileStats (qd->File_weight_approx);
  if (qd->File_text_idx)
    ZeroFileStats (qd->File_text_idx);
}


void 
TransFileStats (query_data * qd)
{
  qd->File_text->Current = qd->File_text->Cumulative;
  if (qd->File_comp_dict)
    qd->File_comp_dict->Current = qd->File_comp_dict->Cumulative;
  if (qd->File_fast_comp_dict)
    qd->File_fast_comp_dict->Current = qd->File_fast_comp_dict->Cumulative;
  qd->File_stem->Current = qd->File_stem->Cumulative;

  /* [RPAP - Jan 97: Stem Index Change] */
  if (qd->File_stem1)
    qd->File_stem1->Current = qd->File_stem1->Cumulative;
  if (qd->File_stem2)
    qd->File_stem2->Current = qd->File_stem2->Cumulative;
  if (qd->File_stem3)
    qd->File_stem3->Current = qd->File_stem3->Cumulative;

  qd->File_invf->Current = qd->File_invf->Cumulative;
  if (qd->File_text_idx_wgt)
    qd->File_text_idx_wgt->Current = qd->File_text_idx_wgt->Cumulative;
  if (qd->File_weight_approx)
    qd->File_weight_approx->Current = qd->File_weight_approx->Cumulative;
  if (qd->File_text_idx)
    qd->File_text_idx->Current = qd->File_text_idx->Cumulative;
}


/**
 * Block: Query Document handling;
 *        Provides basic loading/unloading facilities for individual
 *          documents, usually for presentation purposes (searches
 *          are performed on inverted files
 */

void 
QueryData_FreeTextBuffer (query_data * qd)
{
  if (qd->TextBuffer)
    {
      delete qd->TextBuffer;
      ChangeMemInUse (qd, -qd->TextBufferLen);
    }
  qd->TextBuffer = NULL;
  qd->TextBufferLen = 0;
}

void 
QueryData_FreeQueryDocs (query_data * qd)
{
  qd->doc_pos = 0;
  qd->buf_in_use = 0;
  if (qd->DL)
    {
      //int i;
      DocList_FreeTextBuffers(qd->DL, qd);
      /*
      for (i = 0; i < qd->DL->size(); i++)
	DocEntry_FreeTextBuffer(&qd->DL->DE[i], qd);
        
	if (qd->DL->DE[i].CompTextBuffer)
	  {
	    delete qd->DL->DE[i].CompTextBuffer;
	    qd->DL->DE[i].CompTextBuffer = NULL;
	    ChangeMemInUse (qd, -qd->DL->DE[i].Len);
	  }
      */
      DocList_destroy(qd->DL);
    }
  qd->DL = NULL;
  QueryData_FreeTextBuffer (qd);
}

int 
LoadCompressedText (query_data * qd, int max_mem)
{
  DocEntry *DE;

  if (qd->DL == NULL || qd->doc_pos >= qd->DL->size())
    return -1;

  /**
   *  GRB: added test for td (text data) per addition of loading
   *       indexes only; td may now be NULL; this returns a
   *       special value to  indicated that the text was not
   *       loaded; mgquery etc now test on that value
   *  17/09/99
   */
  if (qd->td == NULL)
    return -2;

  DE = qd->DL->member(qd->doc_pos);

  if (!DocEntry_TextLoaded(DE))
    {
      //int i;
      //DocEntry *de;

      DocList_FreeTextBuffers(qd->DL, qd);
      /*
      for (i = 0, de = qd->DL->DE; i < qd->DL->size(); i++, de++)
	DocEntry_FreeTextBuffer(de);
      */

       if (LoadBuffers (qd, qd->DL->member(qd->doc_pos), max_mem,
		       qd->DL->size() - qd->doc_pos) == -1)
	return -1;
    }
  return 0;
}


int 
GetDocNum (query_data * qd)
{
  if (qd->DL == NULL || qd->doc_pos >= qd->DL->size())
    return -1;
  return qd->DL->member(qd->doc_pos)->docNum();
}

DocEntry *
GetDocChain (query_data * qd)
{
  if (qd->DL == NULL || qd->doc_pos >= qd->DL->size())
    return NULL;
  return qd->DL->member(qd->doc_pos);
}

float 
GetDocWeight (query_data * qd)
{
  if (qd->DL == NULL || qd->doc_pos >= qd->DL->size())
    return -1;
  return qd->DL->member(qd->doc_pos)->docWeight();
}

long 
GetDocCompLength (query_data * qd)
{
  if (qd->DL == NULL || qd->doc_pos >= qd->DL->size())
    return -1;
  return qd->DL->DE[qd->doc_pos].Len;
}


u_char *
GetDocText (query_data * qd, unsigned long *len)
{
  DocEntry *DE;
  int ULen;
  if (qd->DL == NULL || qd->doc_pos >= qd->DL->size())
    return NULL;

  DE = qd->DL->member(qd->doc_pos);

  if (!DocEntry_TextLoaded(DE))
    {
      fprintf (stderr, "The compressed text buffer is NULL\n");
      mg_errno = MG_NOMEM;
      return (NULL);
    }

  QueryData_FreeTextBuffer (qd);

  qd->TextBufferLen = (int) (qd->td->cth.ratio * 1.01 *
			     DocEntry_length(DE)) + 100;
  if (!(qd->TextBuffer = new unsigned char[qd->TextBufferLen]))
    {
      fprintf (stderr, "No memory for TextBuffer\n");
      mg_errno = MG_NOMEM;
      return (NULL);
    }

  DecodeText (qd->cd, (u_char *) DocEntry_TextBuffer(DE), DE->Len,
	      (u_char *) (qd->TextBuffer), &ULen);
  qd->TextBuffer[ULen] = '\0';

  if (ULen >= qd->TextBufferLen)
    {
      fprintf (stderr, "%d >= %d\n", ULen, qd->TextBufferLen);
      mg_errno = MG_BUFTOOSMALL;
      return (NULL);
    }

  if (len)
    *len = ULen;

  return qd->TextBuffer;
}

int 
NextDoc (query_data * qd)
{
  if (qd->DL == NULL || qd->doc_pos >= qd->DL->size())
    return 0;
  qd->doc_pos++;
  return qd->doc_pos < qd->DL->size();
}


